from dataclasses import dataclass, field

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync._internal.canvas_sync_observer import CanvasSyncObserver


@dataclass
class SyncReceipt(CanvasSyncObserver):
    updated_courses: set[str] = field(default_factory=set)
    updated_modules: set[str] = field(default_factory=set)
    updated_pages: set[str] = field(default_factory=set)
    updated_assignment_groups: set[str] = field(default_factory=set)
    updated_assignments: set[str] = field(default_factory=set)
    updated_resources: set[str] = field(default_factory=set)
    updated_messages: set[str | int] = field(default_factory=set)

    def new_content_found(self, content: object):
        if isinstance(content, db.Course):
            self.updated_courses.add(content.id)
        elif isinstance(content, db.Module):
            self.updated_modules.add(content.id)
        elif isinstance(content, db.ModulePage):
            self.updated_pages.add(content.id)
            self.updated_courses.add(content.course_id)
            self.updated_modules.add(content.module_id)
        elif isinstance(content, db.CourseMessage):
            self.updated_messages.add(content.id)
            self.updated_courses.add(content.course_id)
        elif isinstance(content, db.AssignmentGroup):
            self.updated_assignment_groups.add(content.id)
        elif isinstance(content, db.Assignment):
            self.updated_assignments.add(content.id)
            self.updated_assignment_groups.add(content.group_id)
            self.updated_courses.add(content.course_id)
        elif isinstance(content, db.Resource):
            self.updated_resources.add(content.id)
            self.updated_courses.add(content.course_id)



from datetime import datetime
from pathlib import Path

from stones import LmdbStore, stone


class SyncMeta:
    def __init__(self, stone_path: Path):
        self._db = stone(str(stone_path.absolute()), LmdbStore)

    @property
    def last_sync_time(self) -> datetime:
        return self._db.get("last", datetime.min)

    def update_last_sync_time(self):
        self._db["last"] = datetime.now()



import logging
from asyncio import Semaphore
from pathlib import Path
from typing import Generic, Type, TypeVar

from qcanvas_api_clients.canvas import CanvasClient, CanvasClientConfig
from qcanvas_api_clients.panopto import PanoptoClient, PanoptoClientConfig

import qcanvas_backend.database.types as db
from qcanvas_backend.database import QCanvasDatabase
from qcanvas_backend.database.data_monolith import DataMonolith
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.resources.extracting import (
    CanvasFileExtractor,
    CanvasMediaObjectExtractor,
    EmbeddedPanoptoExtractor,
)
from qcanvas_backend.net.resources.extracting.extractors import Extractors
from qcanvas_backend.net.resources.extracting.link_extractor import LinkExtractor
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qcanvas_backend.net.sync.synchronization_client import SynchronizationClient

U = TypeVar("U", bound=ResourceManager)

_logger = logging.getLogger(__name__)


class QCanvas(Generic[U]):
    def __init__(
        self,
        canvas_config: CanvasClientConfig,
        panopto_config: PanoptoClientConfig,
        storage_path: Path,
        resource_manager_class: Type[U],
    ):
        storage_path.mkdir(parents=True, exist_ok=True)

        self._data_cache = None
        self._cache_semaphore = Semaphore()
        self.canvas_client = CanvasClient(client_config=canvas_config)
        self.panopto_client = PanoptoClient(
            client_config=panopto_config, canvas_client=self.canvas_client
        )
        self.database = QCanvasDatabase(storage_path / "qcanvas.db")
        self.resource_manager: U = self._create_resource_manager(
            manager=resource_manager_class,
            database=self.database,
            download_dest=storage_path / "downloads",
            extractors=Extractors(*self._create_extractors()),
        )
        self._synchronization_client = SynchronizationClient(
            database=self.database,
            canvas_client=self.canvas_client,
            panopto_client=self.panopto_client,
            resource_scanner=self.resource_manager.scanner,
            sync_meta_store_location=storage_path / ".sync_meta",
        )
        self._init_called = False

    def _create_extractors(self) -> list[LinkExtractor]:
        return [
            CanvasFileExtractor(self.canvas_client),
            EmbeddedPanoptoExtractor(self.canvas_client, self.panopto_client),
            CanvasMediaObjectExtractor(self.canvas_client),
        ]

    @staticmethod
    def _create_resource_manager(
        manager: Type[U],
        database: QCanvasDatabase,
        download_dest: Path,
        extractors: Extractors,
    ) -> ResourceManager:
        return manager(
            database=database, download_dest=download_dest, extractors=extractors
        )

    async def init(self):
        await self.database.init()
        self.resource_manager.add_existing_resources(
            await self.database.get_existing_resources()
        )

        self._init_called = True

    async def synchronise_canvas(self) -> SyncReceipt:
        assert self._init_called
        _logger.info("Synchronising")

        receipt = await self._synchronization_client.synchronise_canvas()
        await self._reload_data()
        return receipt

    async def _reload_data(self):
        self._data_cache = None
        await self.get_data()

    async def download(self, resource: db.Resource):
        assert self._init_called
        _logger.info(
            "Downloading %s (course_id=%s)", resource.file_name, resource.course_id
        )

        await self.resource_manager.download(resource)

    async def get_data(self) -> DataMonolith:
        assert self._init_called

        async with self._cache_semaphore:
            if self._data_cache is None:
                _logger.info("Loading all data from database")

                self._data_cache = await self.database.get_data()

            return self._data_cache
